import numpy as np
import pandas as pd
import torch
from torch import nn
from torch import optim
from pathlib import Path
from typing import Union, List, Tuple, Collection
from sklearn.model_selection import train_test_split
from torch.nn import functional as F
from torch.optim.lr_scheduler import LRScheduler
from torch.utils.data import DataLoader, Dataset

from nlpx.llm import TokenizeVec
from nlpx.dataset import TokenDataset, PaddingTokenCollator
from nlpx.text_token import BaseTokenizer, PaddingTokenizer, SimpleTokenizer, Tokenizer, TokenEmbedding
from nlpx.training import Trainer, SimpleTrainer, ClassTrainer, SimpleClassTrainer, SplitClassTrainer, evaluate, \
	EvalClassTrainer, EvalSimpleClassTrainer


class ModelWrapper:
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import ModelWrapper
	>>> model_wrapper = ModelWrapper(model)
	>>> model_wrapper.train(train_set, val_set, collate_fn)
	>>> model_wrapper.logits(X_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], device: torch.device = None):
		if isinstance(model_or_path, nn.Module):
			self.model, self.best_model = model_or_path, model_or_path
		elif isinstance(model_or_path, (str, Path)):
			self.model = torch.load(model_or_path, map_location=device)
			self.best_model = self.model
		self.device = device or torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
	
	def train(self, train_set: Dataset, collate_fn=None, max_iter=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          learning_rate=0.001, T_max: int = 0,
	          batch_size=64, num_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		trainer = Trainer(max_iter, optimizer, scheduler, learning_rate, T_max,
		                  batch_size,
		                  num_workers,
		                  pin_memory, pin_memory_device,
		                  persistent_workers,
		                  early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
		                  print_per_rounds,
		                  checkpoint_per_rounds,
		                  checkpoint_name,
		                  self.device)
		self.best_model = trainer.train(self.model, train_set, collate_fn, show_progress, eps)
	
	def logits(self, X: torch.Tensor):
		self.best_model.eval()
		with torch.no_grad():
			logits = self.best_model(X)
		return logits
	
	def save(self, best_model_path: Union[str, Path] = './best_model.pt', 
	         last_model_path: Union[str, Path] = './last_model.pt', save_mode: str = 'both'):
		"""
		:param best_model_path: 
		:param last_model_path: 
		:param save_mode: "both" or "best" or "last"
		:return: 
		"""
		assert save_mode in ('both', 'best', 'last')
		if save_mode == 'both':
			torch.save(self.model, last_model_path)
			torch.save(self.best_model, best_model_path)
		elif save_mode == 'best':
			torch.save(self.best_model, best_model_path)
		elif save_mode == 'last':
			torch.save(self.model, last_model_path)
	
	def load(self, model_path: Union[str, Path] = './best_model.pt'):
		self.model = torch.load(model_path, map_location=self.device)
		self.best_model = self.model


class SimpleModelWrapper(ModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SimpleModelWrapper
	>>> model_wrapper = SimpleModelWrapper(model)
	>>> model_wrapper.train(X, y, collate_fn)
	>>> model_wrapper.logits(X_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             device: torch.device = None):
		super().__init__(model_or_path, device)
	
	def train(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	          collate_fn=None, max_iter=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          learning_rate=0.001, T_max: int = 0,
	          batch_size=64, num_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		trainer = SimpleTrainer(max_iter, optimizer, scheduler, learning_rate, T_max,
		                        batch_size,
		                        num_workers,
		                        pin_memory, pin_memory_device,
		                        persistent_workers,
		                        early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
		                        print_per_rounds,
		                        checkpoint_per_rounds,
		                        checkpoint_name,
		                        self.device)
		self.best_model = trainer.train(self.model, X, y, collate_fn, show_progress, eps)


class ClassModelWrapper(ModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import ClassModelWrapper
	>>> model_wrapper = ClassModelWrapper(model, classes=classes)
	>>> model_wrapper.train(train_set, val_set, collate_fn)
	>>> model_wrapper.predict(X_test)
	>>> model_wrapper.evaluate(test_set)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, device)
		self.classes = classes
	
	def train(self, train_set: Dataset, val_set: Dataset = None, collate_fn=None, max_iter=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          learning_rate=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		if val_set:
			trainer = EvalClassTrainer(max_iter, optimizer, scheduler, learning_rate, T_max,
			                           batch_size, eval_batch_size,
			                           num_workers, num_eval_workers,
			                           pin_memory, pin_memory_device,
			                           persistent_workers,
			                           early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                           print_per_rounds,
			                           checkpoint_per_rounds,
			                           checkpoint_name,
			                           self.device)
			self.best_model = trainer.train(self.model, train_set, val_set, collate_fn, show_progress, eps)
		else:
			trainer = ClassTrainer(max_iter, optimizer, scheduler, learning_rate, T_max,
			                       batch_size, num_workers,
			                       pin_memory, pin_memory_device,
			                       persistent_workers,
			                       early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
			                       print_per_rounds,
			                       checkpoint_per_rounds,
			                       checkpoint_name,
			                       self.device)
			self.best_model = trainer.train(self.model, train_set, collate_fn, show_progress, eps)
	
	def predict(self, X: torch.Tensor):
		logits = self.logits(X)
		return logits.argmax(-1)
	
	def predict_classes(self, X: torch.Tensor):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(X)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]
	
	def predict_proba(self, X: torch.Tensor):
		logits = self.logits(X)
		result = F.softmax(logits, dim=1).max(1)
		return result.indices, result.values.numpy()
	
	def predict_classes_proba(self, X: torch.Tensor):
		assert self.classes is not None, 'classes must be specified'
		indices, values = self.predict_proba(X)
		return [self.classes[i] for i in indices.detach().numpy().ravel()], values
	
	def evaluate(self, val_set: Dataset, batch_size=64, num_workers=0, collate_fn=None):
		val_loader = DataLoader(dataset=val_set, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn)
		_, acc = evaluate(self.best_model, val_loader, self.device)
		return acc


class SimpleClassModelWrapper(ClassModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SimpleClassModelWrapper
	>>> model_wrapper = SimpleClassModelWrapper(model, classes=classes)
	>>> model_wrapper.train(X, y val_data, collate_fn)
	>>> model_wrapper.predict(X_test)
	>>> model_wrapper.evaluate(X_test, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, classes, device)
	
	def train(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	          val_data: Tuple[Union[torch.Tensor, np.ndarray, List], Union[torch.LongTensor, np.ndarray, List]] = None,
	          collate_fn=None, max_iter=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          learning_rate=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		if val_data:
			trainer = EvalSimpleClassTrainer(max_iter, optimizer, scheduler, learning_rate, T_max,
			                                 batch_size, eval_batch_size,
			                                 num_workers, num_eval_workers,
			                                 pin_memory, pin_memory_device,
			                                 persistent_workers,
			                                 early_stopping_rounds,
			                                 print_per_rounds,
			                                 checkpoint_per_rounds,
			                                 checkpoint_name,
			                                 self.device)
			self.best_model = trainer.train(self.model, X, y, val_data, collate_fn, show_progress, eps)
		else:
			trainer = SimpleClassTrainer(max_iter, optimizer, scheduler, learning_rate, T_max,
			                             batch_size, num_workers,
			                             pin_memory, pin_memory_device,
			                             persistent_workers,
			                             early_stopping_rounds,
			                             print_per_rounds,
			                             checkpoint_per_rounds,
			                             checkpoint_name,
			                             self.device)
			self.best_model = trainer.train(self.model, X, y, collate_fn, show_progress, eps)
	
	def evaluate(self, X: Union[torch.Tensor, np.ndarray, List], y: Union[torch.LongTensor, np.ndarray, List],
	             batch_size=64, num_workers=0, collate_fn=None):
		if isinstance(X, (List, np.ndarray)):
			X = torch.tensor(X, dtype=torch.float)
		if isinstance(y, (List, np.ndarray)):
			y = torch.tensor(y, dtype=torch.long)
			
		data_set = TokenDataset(X, y)
		return super().evaluate(data_set, batch_size, num_workers, collate_fn)


class TextModelWrapper(SimpleClassModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import TextModelWrapper
	>>> model_wrapper = TextModelWrapper(model, tokenize_vec, classes=classes)
	>>> model_wrapper.train(train_texts, y_train val_data, collate_fn)
	>>> model_wrapper.predict(test_texts)
	>>> model_wrapper.evaluate(test_texts, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path], 
	             tokenize_vec: Union[BaseTokenizer, TokenizeVec, TokenEmbedding],
	             classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, classes, device)
		self.tokenize_vec = tokenize_vec
	
	def train(self, texts: Union[Collection[str], np.ndarray, pd.Series],
	          y: Union[torch.LongTensor, np.ndarray, List],
	          val_data: Tuple[
		          Union[Collection[str], np.ndarray, pd.Series], Union[torch.LongTensor, np.ndarray, List]] = None,
	          max_length: int = None, collate_fn=None, max_iter=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          learning_rate=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          n_jobs=-1, show_progress=False, eps=1e-5):
		X = self.get_vec(texts, max_length=max_length, n_jobs=n_jobs)
		if val_data:
			val_data = (self.get_vec(val_data[0], max_length=max_length, n_jobs=-1), val_data[1])
		super().train(X, y, val_data, collate_fn, max_iter, optimizer, scheduler, learning_rate, T_max,
		              batch_size, eval_batch_size, num_workers, num_eval_workers, pin_memory, pin_memory_device,
		              persistent_workers, early_stopping_rounds, print_per_rounds, checkpoint_per_rounds,
		              checkpoint_name, show_progress, eps)
	
	def predict(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		logits = self.logits(texts, max_length, n_jobs=n_jobs)
		return logits.argmax(-1)
	
	def predict_classes(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(texts, max_length, n_jobs=n_jobs)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]
	
	def predict_proba(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		logits = self.logits(texts, max_length, n_jobs=n_jobs)
		result = F.softmax(logits, dim=1).max(1)
		return result.indices, result.values.numpy()
	
	def predict_classes_proba(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		assert self.classes is not None, 'classes must be specified'
		indices, values = self.predict_proba(texts, max_length, n_jobs)
		return [self.classes[i] for i in indices.detach().numpy().ravel()], values
	
	def logits(self, texts: Collection[str], max_length: int = None, n_jobs=-1):
		X = self.get_vec(texts, max_length, n_jobs=n_jobs)
		return super().logits(X)
	
	def evaluate(self, texts: Union[str, Collection[str], np.ndarray, pd.Series],
	             y: Union[torch.LongTensor, np.ndarray, List], batch_size=64, num_workers=0,
	             max_length: int = None, collate_fn=None, n_jobs=-1):
		X = self.get_vec(texts, max_length, n_jobs=n_jobs)
		if isinstance(y, (np.ndarray, List)):
			y = torch.tensor(y, dtype=torch.long)
		return super().evaluate(X, y, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn)
	
	def get_vec(self, texts: Union[str, Collection[str], np.ndarray, pd.Series], max_length: int, n_jobs: int):
		if isinstance(texts, str):
			texts = [texts]
		
		if isinstance(self.tokenize_vec, TokenizeVec):
			return self.tokenize_vec.parallel_encode_plus(texts, max_length=max_length, padding='max_length',
			                                              truncation=True, add_special_tokens=True,
			                                              return_token_type_ids=True, return_attention_mask=True,
			                                              return_tensors='pt', n_jobs=n_jobs)
		
		elif isinstance(self.tokenize_vec, (PaddingTokenizer, SimpleTokenizer, Tokenizer)):
			return torch.LongTensor(self.tokenize_vec.batch_encode(texts, max_length))
		
		elif isinstance(self.tokenize_vec, TokenEmbedding):
			return self.tokenize_vec(texts, max_length)
		
		raise ValueError("Invalid tokenize_vec, it must be a TokenizeVec or TokenEmbedding.")


class SplitTextModelWrapper(TextModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SplitTextModelWrapper
	>>> model_wrapper = SplitTextModelWrapper(model, tokenize_vec, classes=classes)
	>>> model_wrapper.train(texts, y, collate_fn)
	>>> model_wrapper.predict(test_texts)
	>>> model_wrapper.evaluate(test_texts, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             tokenize_vec: Union[BaseTokenizer, TokenizeVec, TokenEmbedding],
	             classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, tokenize_vec, classes, device)
	
	def train(self, texts: Union[Collection[str], np.ndarray, pd.Series],
	          y: Union[torch.LongTensor, np.ndarray, List],
	          max_length: int = None, eval=True, val_size=0.2, random_state=None, collate_fn=None, max_iter=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          learning_rate=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          n_jobs=-1, show_progress=False, eps=1e-5):
		X = self.get_vec(texts, max_length=max_length, n_jobs=n_jobs)
		if eval:
			trainer = SplitClassTrainer(max_iter, optimizer, scheduler, learning_rate, T_max,
			                            batch_size, eval_batch_size,
			                            num_workers, num_eval_workers,
			                            pin_memory, pin_memory_device,
			                            persistent_workers,
			                            early_stopping_rounds,
			                            print_per_rounds,
			                            checkpoint_per_rounds,
			                            checkpoint_name,
			                            self.device)
			self.best_model = trainer.train(self.model, X, y, val_size, random_state, collate_fn, show_progress, eps)
		else:
			trainer = SimpleClassTrainer(max_iter, optimizer, scheduler, learning_rate, T_max,
			                             batch_size, num_workers,
			                             pin_memory, pin_memory_device,
			                             persistent_workers,
			                             early_stopping_rounds,
			                             print_per_rounds,
			                             checkpoint_per_rounds,
			                             checkpoint_name,
			                             self.device)
			self.best_model = trainer.train(self.model, X, y, collate_fn, show_progress, eps)


class PaddingTextModelWrapper(ClassModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import PaddingTextModelWrapper
	>>> model_wrapper = PaddingTextModelWrapper(model, tokenizer, classes=classes)
	>>> model_wrapper.train(train_texts, y_train val_data)
	>>> model_wrapper.predict(test_texts)
	>>> model_wrapper.evaluate(test_texts, y_test)
	"""
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             tokenizer: Union[PaddingTokenizer, SimpleTokenizer, Tokenizer],
	             classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, classes, device)
		self.tokenizer = tokenizer
	
	def train(self, texts: Union[Collection[str], np.ndarray, pd.Series],
	          y: Union[torch.LongTensor, np.ndarray, List],
	          val_data: Tuple[Union[Collection[str], np.ndarray, pd.Series], Union[torch.LongTensor, np.ndarray, List]] = None,
	          max_length: int = None, max_iter=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          learning_rate=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		X = self.tokenizer.batch_encode(texts, padding=False)
		train_set = TokenDataset(X, y)
		val_set = None
		if val_data:
			X_val = self.tokenizer.batch_encode(val_data[0], padding=False)
			val_set = TokenDataset(X_val, val_data[1])
		
		super().train(self.model, train_set, val_set, collate_fn=PaddingTokenCollator(self.tokenizer.pad, max_length),
		              max_iter=max_iter, optimizer=optimizer, scheduler=scheduler, learning_rate=learning_rate,
		              T_max=T_max, batch_size=batch_size, num_workers=num_workers, pin_memory=pin_memory,
		              pin_memory_device=pin_memory_device, persistent_workers=persistent_workers,
		              early_stopping_rounds=early_stopping_rounds, print_per_rounds=print_per_rounds,
		              checkpoint_per_rounds=checkpoint_per_rounds, checkpoint_name=checkpoint_name,
		              show_progress=show_progress, eps=eps)
	
	def predict(self, texts: Collection[str], max_length: int = None):
		logits = self.logits(texts, max_length)
		return logits.argmax(-1)
	
	def predict_classes(self, texts: Collection[str], max_length: int = None):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(texts, max_length)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]
	
	def predict_proba(self, texts: Collection[str], max_length: int = None):
		logits = self.logits(texts, max_length)
		result = F.softmax(logits, dim=1).max(1)
		return result.indices, result.values.numpy()
	
	def predict_classes_proba(self, texts: Collection[str], max_length: int = None):
		assert self.classes is not None, 'classes must be specified'
		indices, values = self.predict_proba(texts, max_length)
		return [self.classes[i] for i in indices.detach().numpy().ravel()], values
	
	def logits(self, texts: Collection[str], max_length: int = None):
		X = self.tokenizer.batch_encode(texts, max_length)
		X = torch.tensor(X, dtype=torch.long)
		return super().logits(X)
	
	def evaluate(self, texts: Union[str, Collection[str], np.ndarray, pd.Series],
	             y: Union[torch.LongTensor, np.ndarray, List], batch_size=64, num_workers=0,
	             max_length: int = None):
		X = self.tokenizer.batch_encode(texts, padding=False)
		if isinstance(y, (np.ndarray, List)):
			y = torch.tensor(y, dtype=torch.long)
		val_set = TokenDataset(X, y)
		return super().evaluate(val_set, batch_size=batch_size, num_workers=num_workers,
		                        collate_fn=PaddingTokenCollator(self.tokenizer.pad, max_length))


class SplitPaddingTextModelWrapper(PaddingTextModelWrapper):
	"""
	Examples
	--------
	>>> from nlpx.model.wrapper import SplitPaddingTextModelWrapper
	>>> model_wrapper = SplitPaddingTextModelWrapper(tokenizer, classes=classes)
	>>> model_wrapper.train(model, texts, y)
	>>> model_wrapper.predict(test_texts)
	>>> model_wrapper.evaluate(test_texts, y_test)
	"""
	
	def __init__(self, model_or_path: Union[nn.Module, str, Path],
	             tokenizer: Union[PaddingTokenizer, SimpleTokenizer, Tokenizer],
	             classes: Collection[str] = None,
	             device: torch.device = None):
		super().__init__(model_or_path, tokenizer, classes, device)
	
	def train(self, texts: Union[Collection[str], np.ndarray, pd.Series],
	          y: Union[torch.LongTensor, np.ndarray, List],
	          max_length: int = None, eval=True, val_size=0.2, random_state=None, max_iter=100,
	          optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	          learning_rate=0.001, T_max: int = 0,
	          batch_size=64, eval_batch_size=128,
	          num_workers=0, num_eval_workers=0,
	          pin_memory: bool = False, pin_memory_device: str = "",
	          persistent_workers: bool = False,
	          early_stopping_rounds: int = 10,
	          print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt',
	          show_progress=False, eps=1e-5):
		val_data = None
		if eval:
			X_train, X_test, y_train, y_test = train_test_split(texts, y, test_size=val_size, random_state=random_state)
			val_data = (X_test, y_test)
		else:
			X_train, y_train = texts, y
		
		super().train(X_train, y_train, val_data, max_length, max_iter, optimizer, scheduler, learning_rate,
		              T_max, batch_size, eval_batch_size, num_workers, num_eval_workers, pin_memory, pin_memory_device,
		              persistent_workers, early_stopping_rounds, print_per_rounds, checkpoint_per_rounds,
		              checkpoint_name, show_progress, eps)



import gc
import pickle
import torch
from torch.nn import Embedding
from pathlib import Path
from itertools import tee
from typing import Union, List, Collection

from .utils import UTF8, read_file, read_corpus_files, batch_cut, batch_pad_mask, cut, batch_pad, pad, \
    filter_stop_words, pad_mask, load_embedding
from .stats import token_counter


class BaseTokenizer:
    """
    如果碰到未登录词会报错，不能padding，记住 min_freq = 0，否则易报错；如果没有特殊要求，请使用其子类SimpleTokenizer
    """
    # SAVE_SEP = '\t'
    RESERVED_TOKENS = []
    
    def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
                 cut_texts: Union[Collection[Collection[str]], Collection[str]] = None, vocab: Collection[str] = None,
                 min_freq=0, special_tokens: Collection[str] = None, lang='cn', cut_type='word', word_freq: bool = False,
                 stop_words: Collection[str] = None, keep_punctuation=False, cut_fn=None):
        """
        :param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
        :param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        self.lang = lang
        self.cut_type = cut_type
        self.stop_words = stop_words,
        self.keep_punctuation = keep_punctuation
        if cut_fn is None and lang == 'cn' and cut_type == 'word':
            import jieba
            import logging
            jieba.setLogLevel(logging.INFO)
            self.cut_fn = jieba.lcut
        else:
            self.cut_fn = cut_fn
            
        self.special_tokens = self.RESERVED_TOKENS
        if special_tokens:
            self.special_tokens.extend([token for token in special_tokens if token not in self.RESERVED_TOKENS])
            
        if vocab:
            if stop_words:
                for word in stop_words:
                    if word in vocab:
                        vocab.remove(word)
            
            if cut_fn is None and lang == 'cn' and cut_type == 'word':
                for word in vocab:
                    jieba.add_word(word)
            
            for token in {token for token in self.special_tokens if token in vocab}:
                vocab.remove(token)
            
            self.vocab = self.special_tokens + vocab if isinstance(vocab, List) else list(vocab)
            self.word_to_idx = {k: i for i, k in enumerate(self.vocab)}
            del vocab
            gc.collect()
        
        else:
            if path is not None and texts is None and cut_texts is None:
                texts = read_corpus_files(path)
                
            if texts is not None:
                cut_texts = batch_cut(texts, lang=lang, cut_type=cut_type, keep_punctuation=keep_punctuation,
                                      cut_fn=self.cut_fn)
                del texts
                
            if cut_texts is not None:
                if isinstance(cut_texts, Collection) and isinstance(cut_texts[0], str):
                    cut_texts = batch_cut(cut_texts, lang='en', keep_punctuation=keep_punctuation)
                if stop_words:
                    cut_texts = filter_stop_words(cut_texts, stop_words)
                counter = token_counter(cut_texts)
                sorted_token_freq = sorted(counter.items(), key=lambda kv: kv[1], reverse=True)
                self.vocab = self.special_tokens.copy()
                if min_freq > 0:
                    filter_tokens = filter(lambda kv: kv[1] >= min_freq and kv[0] not in self.vocab,
                                           sorted_token_freq)
                else:
                    filter_tokens = filter(lambda kv: kv[0] not in self.vocab, sorted_token_freq)
                
                if word_freq:
                    filter_tokens, filter_tokens_copy = tee(filter_tokens, 2)
                    self.word_freq = list(filter_tokens_copy)
                self.vocab += list(map(lambda kv: kv[0], filter_tokens))
                self.word_to_idx = {k: i for i, k in enumerate(self.vocab)}
                del sorted_token_freq, counter, cut_texts
                gc.collect()
            
            else:
                raise ValueError('参数file, texts, vocab不能同时为None.')

        self.special_token_ids = list(map(self._do_encode, self.special_tokens))
    
    @classmethod
    def from_file(cls, path: Union[str, Path], encoding=UTF8, pattern='*', func=read_file, min_freq=0,
                  special_tokens: Collection[str] = None, lang='cn', cut_type='word', word_freq=False,
                  stop_words: Collection[str] = None, keep_punctuation=False, cut_fn=None):
        """
        :param path: 语料文件，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param encoding: 编码
        :param pattern: 文件后缀，当file是文件夹的时候，会根据此后缀过滤文件
        :param func: 具体读取文件的处理函数，默认是read_file，可替换。注意：其函数签名为 function_name(path: str, encoding: str) -> texts: Collection[str]
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        texts = read_corpus_files(path, encoding, pattern, func)
        return cls(texts=texts, min_freq=min_freq, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
                   word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
    
    @classmethod
    def from_texts(cls, texts: Collection[str], min_freq=0, special_tokens: Collection[str] = None, lang='cn',
                    cut_type='word', word_freq=False, stop_words: Collection[str] = None, keep_punctuation=False,
                   cut_fn=None):
        """
        :param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        return cls(texts=texts, min_freq=min_freq, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
                   word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
    
    @classmethod
    def from_cut_texts(cls, cut_texts: Collection[Collection[str]], min_freq=0, special_tokens: Collection[str] = None,
                        lang='cn', cut_type='word', word_freq=False, stop_words: Collection[str] = None,
                        keep_punctuation=False, cut_fn=None):
        """
        :param cut_texts: 分词后的语料，每个元素是一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        return cls(cut_texts=cut_texts, min_freq=min_freq, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
                   word_freq=word_freq, stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
    
    @classmethod
    def from_vocab(cls, vocab: Collection[str], special_tokens: Collection[str] = None, lang='cn', cut_type='word',
                   stop_words: Collection[str] = None, keep_punctuation=False, cut_fn=None):
        """
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        return cls(vocab=vocab, special_tokens=special_tokens, lang=lang, cut_type=cut_type, stop_words=stop_words,
                   keep_punctuation=keep_punctuation, cut_fn=cut_fn)
    
    @property
    def vocab_size(self):
        return len(self.vocab)
    
    @property
    def real_vocab(self):
        """
        :return: 除去特殊字符的词表
        """
        special_token_len = len(self.special_tokens)
        idx = special_token_len - 1
        if self.special_tokens[idx] == self.vocab[idx]:
            return self.vocab[len(self.special_tokens):]
        
        for i in range(idx):
            if self.special_tokens[i] != self.vocab[i]:
                idx = i
                break
        return self.vocab[idx:len(self.vocab) + idx - special_token_len]
    
    def encode(self, sentence: Union[str, Collection[str]], max_length: int = None, truncation=True,
               is_split_into_words: bool = False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
        :param max_length:
        :param truncation:
        :param is_split_into_words: 是否已经分词
        :return:
        """
        tokens = self._encode(sentence, is_split_into_words)
        if truncation and max_length:
            return tokens[:max_length]
        return tokens
        
    def _encode(self, sentence: Union[str, Collection[str]], is_split_into_words):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if is_split_into_words:
            if isinstance(sentence, str):
                words = cut(sentence, lang='en', keep_punctuation=self.keep_punctuation)
            elif isinstance(sentence, Collection):
                words = sentence
        else:
            if isinstance(sentence, str):
                words = cut(sentence, self.lang, self.cut_type, keep_punctuation=self.keep_punctuation, cut_fn=self.cut_fn)
            else:
                raise ValueError("当参数is_split_into_words为False时，参数'sentence'只能是str类型")

        if self.stop_words:
            return self.do_encode_with_stop_words(words)
        else:
            return self.do_encode(words)
    
    def batch_encode(self, sentences: Union[str, Collection[str], Collection[Collection[str]]], max_length: int = None,
                     truncation=True, is_split_into_words: bool = False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
            return self.encode(sentences, max_length, truncation, is_split_into_words)
        
        tokens = self._batch_encode(sentences, is_split_into_words)
        if truncation and max_length:
            return [x[:max_length] for x in tokens]
        return list(tokens)
    
    def _batch_encode(self, sentences: Union[Collection[str], Collection[Collection[str]]], is_split_into_words):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if isinstance(sentences, Collection):
            if is_split_into_words:
                if self.stop_words:
                    return map(self.do_encode_with_stop_words, sentences)
                return map(self.do_encode, sentences)
            else:
                batch_cuts = batch_cut(sentences, lang=self.lang, cut_type=self.cut_type,
                                       keep_punctuation=self.keep_punctuation, cut_fn=self.cut_fn)
                if self.stop_words:
                    return map(self.do_encode_with_stop_words, batch_cuts)
                return map(self.do_encode, batch_cuts)
        
        raise ValueError('参数"sentence"类型错误')

    def __call__(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None, truncation=True,
                     is_split_into_words: bool = False):
        return self.batch_encode(sentences, max_length, truncation, is_split_into_words)
    
    def do_encode(self, words: Union[str, Collection[str]]):
        """
        把词转换成数字
        :param words: '学生' 或 ['学生', '手机', '老师']
        :return:
        """
        if isinstance(words, str):
            return self._do_encode(words)
        return list(filter(lambda x: x is not None, map(self._do_encode, words)))
    
    def do_encode_with_stop_words(self, words: Union[str, Collection[str]]):
        """
        把词转换成数字
        :param words: '学生' 或 ['学生', '手机', '老师']
        :return:
        """
        if isinstance(words, str):
            return self._do_encode_with_stop_words(words)
        return list(filter(lambda x: x is not None, map(self._do_encode_with_stop_words, words)))
        
    def _do_encode(self, word: str):
        return self.word_to_idx[word]
    
    def _do_encode_with_stop_words(self, cut_token: str):
        return self.word_to_idx[cut_token] if cut_token not in self.stop_words else None
    
    def decode(self, tokens: Union[int, Collection[int]], return_special_tokens=False, return_sentence=False):
        """
        :param tokens: [2, 19, 27, 3, 0, 0]
        :param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
        :param return_sentence: 返回的是一句话还是词序列
        :return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
        """
        if isinstance(tokens, int):
            return self.vocab[tokens]
            
        if return_sentence:
            return ''.join(self._decode(tokens, return_special_tokens))
        return list(self._decode(tokens, return_special_tokens))
    
    def batch_decode(self, batch_tokens: Collection[Collection[int]], return_special_tokens=False, return_sentence=False):
        """
        :param batch_tokens: [[2, 19, 27, 3, 0, 0], [2, 10, 3, 0, 0, 0]]
        :param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
        :param return_sentence: 返回的是一句话还是词序列
        :return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
        """
        if isinstance(batch_tokens[0], int):
            return self.decode(batch_tokens, return_special_tokens, return_sentence)
        return list(map(lambda x: self.decode(x, return_special_tokens, return_sentence), batch_tokens))
    
    def add_special_tokens(self, special_tokens: Collection[str]):
        for token in special_tokens:
            if token in self.special_tokens:
                continue
            self.special_tokens.append(token)
            self.word_to_idx[token] = len(self.vocab)
            if token not in self.vocab:
                self.vocab.append(token)
        self.special_token_ids = list(map(self._do_encode, self.special_tokens))

    def add_words(self, words: Union[str, Collection[str]]):
        if isinstance(words, str):
            self.word_to_idx[words] = len(self.vocab)
            self.vocab.append(words)
        else:
            for word in words:
                if word in self.vocab:
                    continue
                self.word_to_idx[word] = len(self.vocab)
                self.vocab.append(word)
    
    def add_stop_words(self, stop_words: Collection[str]):
        if not stop_words:
            return
        
        if self.stop_words is None:
            self.stop_words = stop_words if isinstance(self.stop_words, List) else list(stop_words)
            for word in stop_words:
                if word in self.vocab:
                    self.vocab.remove(word)
        else:
            if not isinstance(self.stop_words, List):
                self.stop_words = list(self.stop_words)
            
            for word in stop_words:
                if word in self.stop_words:
                    continue
                self.stop_words.append(word)
                if word in self.vocab:
                    self.vocab.remove(word)
        
        self.word_to_idx = {k: i for i, k in enumerate(self.vocab)}
    
    def save(self, path: str = None):
        path = path or f'{self.__class__.__name__}.bin'
        obj = {
            'name': f'{self.__class__.__name__}',
            'lang': self.lang,
            'cut_type': self.cut_type,
            'keep_punctuation': self.keep_punctuation,
            'special_tokens': self.special_tokens,
            'stop_words': self.stop_words,
            'vocab': self.vocab
        }
        with open(path, 'wb') as f:
            pickle.dump(obj, f)

    @classmethod
    def load(cls, path: str = None, cut_fn=None):
        path = path or f'{cls.__name__}.bin'
        with open(path, 'rb') as f:
            obj = pickle.load(f)
        
        tokenizer_name = obj['name']
        assert cls.__name__ == tokenizer_name, f"Expected {tokenizer_name}, but found {cls.__name__}"
        return cls(vocab=obj['vocab'], special_tokens=obj['special_tokens'], lang=obj['lang'], cut_type=obj['cut_type'],
                   stop_words=obj['stop_words'], keep_punctuation=obj['keep_punctuation'], cut_fn=cut_fn)
    
    def _decode(self, tokens: Collection[int], return_special_tokens):
        if return_special_tokens:
            return map(lambda i: self.vocab[i], tokens)
        
        if self.special_tokens:
            return map(lambda i: self.vocab[i], filter(lambda i: i not in self.special_token_ids, tokens))
        
        return map(lambda i: self.vocab[i], tokens)
    
    def __len__(self):
        return self.vocab_size
    
    def __getitem__(self, index):
        return self.vocab[index]


class PaddingTokenizer(BaseTokenizer):
    """
    如果碰到未登录词会用'<unk>'代替，能padding, '<pad>'和'<unk>'相同
    """
    UNK = '<UNK>'
    PAD_ID = UNK_ID = 0
    RESERVED_TOKENS = [UNK]
    
    def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None, cut_texts: Collection[Collection[str]] = None,
                 vocab: Collection[str] = None, min_freq=10, special_tokens: Collection[str] = None, lang='cn',
                 cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
                 cut_fn=None):
        """
        :param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
        :param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
                         special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
                         stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
    
    def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
               is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
        """
        tokens = self._encode(sentence, is_split_into_words)
        if padding:
            return self.pad(tokens, max_length, truncation, padding_side, return_sequence_length)
        
        if truncation and max_length:
            tokens = tokens[:max_length]
        
        if return_sequence_length:
            return tokens, len(tokens)
        return tokens
        
    def batch_encode(self, sentences: Union[Collection[str], Collection[List]], max_length: int = None, truncation=True,
                     padding=True, padding_side='right', is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
            
            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
                
        """
        if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
            return self.encode(sentences, max_length, truncation, padding, padding_side,
                               is_split_into_words, return_sequence_length)

        tokens = self._batch_encode(sentences, is_split_into_words)
        if padding:
            return self.pad(tokens, max_length, truncation, padding_side, return_sequence_length)
        
        tokens = list(tokens)
        if truncation and max_length:
            tokens = [x[:max_length] for x in tokens]
            
        if return_sequence_length:
            return tokens, [len(x) for x in tokens]
        return tokens
        
    def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None, truncation=True,
            padding_side='right', return_sequence_length=False):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding_side:
        :param return_sequence_length: 是否返回序列实际长度
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
            
            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
                
        """
        if isinstance(tokens, map) or isinstance(tokens[0], Collection):
            if return_sequence_length:
                return batch_pad(tokens, max_length, truncation, True, padding_side, False, False, self.PAD_ID, None, None)
            return batch_pad(tokens, max_length, truncation, True, padding_side, False, False, self.PAD_ID, None, None)[0]

        if return_sequence_length:
            return pad(tokens, max_length, truncation, True, padding_side, False, False, self.PAD_ID, None, None)
        return pad(tokens, max_length, truncation, True, padding_side, False, False, self.PAD_ID, None, None)[0]

    def __call__(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None, truncation=True,
                     padding=True, padding_side='right', is_split_into_words: bool = False,
                     return_sequence_length=False):
        return self.batch_encode(sentences, max_length, truncation, padding, padding_side, is_split_into_words,
                                 return_sequence_length)
    
    def _do_encode(self, cut_token: str):
        return self.word_to_idx.get(cut_token, self.UNK_ID)
    
    def _do_encode_with_stop_words(self, cut_token: str):
        return self.word_to_idx.get(cut_token, self.UNK_ID) if cut_token not in self.stop_words else None


class SimpleTokenizer(PaddingTokenizer):
    """
    如果碰到未登录词会用'<unk>'代替，能padding, '<pad>'和'<unk>'不同
    """
    PAD, UNK = '<PAD>', '<UNK>'
    PAD_ID, UNK_ID = 0, 1
    RESERVED_TOKENS = [PAD, UNK]
    
    def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
                 cut_texts: Collection[Collection[str]] = None,
                 vocab: Collection[str] = None, min_freq=5, special_tokens: Collection[str] = None, lang='cn',
                 cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
                 cut_fn=None):
        """
        :param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
        :param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
                         special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
                         stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)


class Tokenizer(PaddingTokenizer):
    """
    如果碰到未登录词会用'<unk>'代替，能padding, 有encode_plus, batch_encode_plus, 返回 {'input_tokens': list, 'mask_tokens': list}
    """
    PAD, UNK, BOS, EOS = '<PAD>', '<UNK>', '<BOS>', '<EOS>'
    PAD_ID, UNK_ID, BOS_ID, EOS_ID = 0, 1, 2, 3
    RESERVED_TOKENS = [PAD, UNK, BOS, EOS]
    
    def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None, cut_texts: Collection[Collection[str]] = None,
                 vocab: Collection[str] = None, min_freq=10, special_tokens: Collection[str] = None, lang='cn',
                 cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
                 cut_fn=None):
        """
        :param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
        :param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
                         special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
                         stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
    
    def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
               bos=True, eos=True, is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
        """
        tokens = self._encode(sentence, is_split_into_words)
        if padding or bos or eos:
            return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, bos, eos)
        
        if truncation and max_length:
            tokens = tokens[:max_length]
            
        if return_sequence_length:
            return tokens, len(tokens)
        return tokens
        
    def batch_encode(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
                     truncation=True, padding=True, padding_side='right', bos=True, eos=True,
                     is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
            
            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
        """
        if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
            return self.encode(sentences, max_length, truncation, padding, padding_side, bos, eos,
                               is_split_into_words, return_sequence_length)

        tokens = self._batch_encode(sentences, is_split_into_words)
        if padding or bos or eos:
            return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, bos, eos)
        
        tokens = list(tokens)
        if truncation and max_length:
            tokens = [x[:max_length] for x in tokens]

        if return_sequence_length:
            return tokens, [len(x) for x in tokens]
        return tokens
        
    def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
            truncation=True, padding=True, padding_side='right', return_sequence_length=False, bos=True, eos=True):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding
        :param padding_side:
        :param return_sequence_length: 是否返回序列实际长度
        :param bos:
        :param eos:
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
            
            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
        """
        if isinstance(tokens, map) or isinstance(tokens[0], Collection):
            if return_sequence_length:
                return batch_pad(tokens, max_length, truncation, padding, padding_side, bos, eos, self.PAD_ID,
                                 self.BOS_ID, self.EOS_ID)
            return batch_pad(tokens, max_length, truncation, padding, padding_side, bos, eos, self.PAD_ID,
                             self.BOS_ID, self.EOS_ID)[0]
        if return_sequence_length:
            return pad(tokens, max_length, truncation, padding, padding_side, bos, eos, self.PAD_ID, self.BOS_ID, self.EOS_ID)
        return pad(tokens, max_length, truncation, padding, padding_side, bos, eos, self.PAD_ID, self.BOS_ID, self.EOS_ID)[0]

    def encode_plus(self, sentence: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
                    truncation=True, padding=True, padding_side='right', bos=False, eos=False,
                    return_mask=False, is_split_into_words=False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了' 或 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param return_mask: 是否返回 mask_tokens
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if isinstance(sentence, str):
            tokens = self._encode(sentence, is_split_into_words)
            return self.pad_plus(tokens, max_length, truncation, padding, padding_side, bos, eos, return_mask)

        raise ValueError('参数"sentence"类型错误')

    def batch_encode_plus(self, sentences: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
                          truncation=True, padding=True, padding_side='right', bos=False, eos=False,
                          return_mask=False, is_split_into_words=False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param return_mask: 是否返回 mask_tokens
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
            return self.encode_plus(sentences, max_length, truncation, padding, padding_side, bos, eos,
                                    return_mask, is_split_into_words)
        if isinstance(sentences, Collection):
            tokens = self._batch_encode(sentences, is_split_into_words)
            return self.pad_plus(tokens, max_length, truncation, padding, padding_side, bos, eos, return_mask)

        raise ValueError('参数"sentence"类型错误')

    def pad_plus(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
                 truncation=True, padding=True, padding_side='right', bos=False, eos=False, return_mask=False):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding
        :param padding_side:
        :param bos:
        :param eos:
        :param return_mask:
        :return:
        """
        if isinstance(tokens, map) or isinstance(tokens[0], Collection):
            if return_mask:
                return batch_pad_mask(tokens, max_length, truncation, padding, padding_side, bos, eos,
                                      self.PAD_ID, self.BOS_ID, self.EOS_ID)
            return {'input_tokens': batch_pad(tokens, max_length, truncation, padding, padding_side, bos, eos,
                                           self.PAD_ID, self.BOS_ID, self.EOS_ID)[0]}

        if return_mask:
            return pad_mask(tokens, max_length, truncation, padding, padding_side, bos, eos,
                            self.PAD_ID, self.BOS_ID, self.EOS_ID)
        return {'input_tokens': pad(tokens, max_length, truncation, padding, padding_side, bos, eos,
                                 self.PAD_ID, self.BOS_ID, self.EOS_ID)[0]}

    def __call__(self, sentence: Union[str, Collection[str]], max_length: int = None, truncation=True, padding=True,
                 padding_side='right', bos=False, eos=False, return_mask=False,
                 is_split_into_words=False):
        return self.encode_plus(sentence, max_length, truncation, padding, padding_side, bos, eos, return_mask,
                                is_split_into_words)


class BosTokenizer(PaddingTokenizer):
    """
    如果碰到未登录词会用'<unk>'代替，能padding, 只有BOS没有EOS
    """
    PAD, UNK, BOS = '<PAD>', '<UNK>', '<BOS>'
    PAD_ID, UNK_ID, BOS_ID = 0, 1, 2
    RESERVED_TOKENS = [PAD, UNK, BOS]
    
    def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
                 cut_texts: Collection[Collection[str]] = None,
                 vocab: Collection[str] = None, min_freq=10, special_tokens: Collection[str] = None, lang='cn',
                 cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
                 cut_fn=None):
        """
        :param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
        :param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
                         special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
                         stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
    
    def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
               bos=True, is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
        """
        tokens = self._encode(sentence, is_split_into_words)
        if padding or bos:
            return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, bos)
        
        if truncation and max_length:
            tokens = tokens[:max_length]
        
        if return_sequence_length:
            return tokens, len(tokens)
        return tokens
    
    def batch_encode(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
                     truncation=True, padding=True, padding_side='right', bos=True,
                     is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)

            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
        """
        if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
            return self.encode(sentences, max_length, truncation, padding, padding_side, bos,
                               is_split_into_words, return_sequence_length)
        
        tokens = self._batch_encode(sentences, is_split_into_words)
        if padding or bos:
            return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, bos)
        
        tokens = list(tokens)
        if truncation and max_length:
            tokens = [x[:max_length] for x in tokens]
        
        if return_sequence_length:
            return tokens, [len(x) for x in tokens]
        return tokens
    
    def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
            truncation=True, padding=True, padding_side='right', return_sequence_length=False, bos=True):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding
        :param padding_side:
        :param return_sequence_length: 是否返回序列实际长度
        :param bos:
        :param eos:
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)

            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
        """
        if isinstance(tokens, map) or isinstance(tokens[0], Collection):
            if return_sequence_length:
                return batch_pad(tokens, max_length, truncation, padding, padding_side, bos, False, self.PAD_ID, self.BOS_ID)
            return batch_pad(tokens, max_length, truncation, padding, padding_side, bos, False, self.PAD_ID, self.BOS_ID)[0]
        if return_sequence_length:
            return pad(tokens, max_length, truncation, padding, padding_side, bos, False, self.PAD_ID, self.BOS_ID)
        return pad(tokens, max_length, truncation, padding, padding_side, bos, False, self.PAD_ID, self.BOS_ID)[0]


class EosTokenizer(PaddingTokenizer):
    """
    如果碰到未登录词会用'<unk>'代替，能padding,  只有EOS没有BOS
    """
    PAD, UNK, EOS = '<PAD>', '<UNK>', '<EOS>'
    PAD_ID, UNK_ID, EOS_ID = 0, 1, 2
    RESERVED_TOKENS = [PAD, UNK, EOS]
    
    def __init__(self, path: Union[str, Path] = None, texts: Collection[str] = None,
                 cut_texts: Collection[Collection[str]] = None,
                 vocab: Collection[str] = None, min_freq=10, special_tokens: Collection[str] = None, lang='cn',
                 cut_type='word', word_freq: bool = False, stop_words: Collection[str] = None, keep_punctuation=False,
                 cut_fn=None):
        """
        :param path: 语料文件路径，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param texts: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
        :param cut_texts: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param min_freq: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_freq: 是否统计词频
        :param stop_words: 停用词
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        super().__init__(path=path, texts=texts, cut_texts=cut_texts, vocab=vocab, min_freq=min_freq,
                         special_tokens=special_tokens, lang=lang, cut_type=cut_type, word_freq=word_freq,
                         stop_words=stop_words, keep_punctuation=keep_punctuation, cut_fn=cut_fn)
    
    def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
               eos=True, is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param eos:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
        """
        tokens = self._encode(sentence, is_split_into_words)
        if padding or eos:
            return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, eos)
        
        if truncation and max_length:
            tokens = tokens[:max_length]
        
        if return_sequence_length:
            return tokens, len(tokens)
        return tokens
    
    def batch_encode(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None,
                     truncation=True, padding=True, padding_side='right', eos=True,
                     is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param eos:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)

            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
        """
        if isinstance(sentences, str) or (is_split_into_words and isinstance(sentences[0], str)):
            return self.encode(sentences, max_length, truncation, padding, padding_side, eos,
                               is_split_into_words, return_sequence_length)
        
        tokens = self._batch_encode(sentences, is_split_into_words)
        if padding or eos:
            return self.pad(tokens, max_length, truncation, padding, padding_side, return_sequence_length, eos)
        
        tokens = list(tokens)
        if truncation and max_length:
            tokens = [x[:max_length] for x in tokens]
        
        if return_sequence_length:
            return tokens, [len(x) for x in tokens]
        return tokens
    
    def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int = None,
            truncation=True, padding=True, padding_side='right', return_sequence_length=False, eos=True):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding
        :param padding_side:
        :param return_sequence_length: 是否返回序列实际长度
        :param eos:
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)

            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
        """
        if isinstance(tokens, map) or isinstance(tokens[0], Collection):
            if return_sequence_length:
                return batch_pad(tokens, max_length, truncation, padding, padding_side, False, eos, self.PAD_ID,
                                 None, self.EOS_ID)
            return batch_pad(tokens, max_length, truncation, padding, padding_side, False, eos, self.PAD_ID,
                             None, self.EOS_ID)[0]
        if return_sequence_length:
            return pad(tokens, max_length, truncation, padding, padding_side, False, eos, self.PAD_ID, None, self.EOS_ID)
        return pad(tokens, max_length, truncation, padding, padding_side, False, eos, self.PAD_ID, None, self.EOS_ID)[0]


class AutoTokenizer:
    TOKENIZER_DICT = {
        "BaseTokenizer": BaseTokenizer,
        "PaddingTokenizer": PaddingTokenizer,
        "SimpleTokenizer": SimpleTokenizer,
        "Tokenizer": Tokenizer,
        "BosTokenizer": BosTokenizer,
        "EosTokenizer": EosTokenizer,
    }
    
    @staticmethod
    def load(path, cut_fn=None):
        with open(path, 'rb') as f:
            obj = pickle.load(f)
    
        tokenizer_name = obj['name']
        tokenizer = AutoTokenizer.TOKENIZER_DICT.get(tokenizer_name)
        assert tokenizer is not None, f"{tokenizer_name} is not a valid tokenizer"
        return tokenizer(vocab=obj['vocab'], special_tokens=obj['special_tokens'], lang=obj['lang'],
                         cut_type=obj['cut_type'], stop_words=obj['stop_words'],
                         keep_punctuation=obj['keep_punctuation'], cut_fn=cut_fn)


class TokenEmbedding:
    """
    可以传入已经训练好的embedding文件路径，也可以embedding数据, encode返回的是 {'input_tokens': list} 或 {'input_tokens': list, 'mask_tokens': list}
    """

    def __init__(self, file: Union[str, Path] = None, vocab: Collection[str] = None, embedding: Collection[Collection[float]] = None,
                 special_tokens: Collection[str] = None, lang='cn', cut_type='word', func=load_embedding,
                 is_large_file=False, keep_punctuation=True, cut_fn=None):
        """
        :param file: embedding文件路径， 如：'./sgns.weibo.word.bz2'
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']，与embedding必须同时传入
        :param embedding: [[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578]]与vocab必须同时传入
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param func: 具体读取文件的处理函数，load_embedding，可替换。
               注意：其函数签名为 function_name(path: str, is_large_file: bool) -> (vocab: list[str], embedding: list[list[float]])
        :param is_large_file: 是否是大文件
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        if file:
            if isinstance(file, str):
                file = Path(file)
            assert file.is_file(), 'file必须是具体文件,不能是文件夹'
            vocab, embedding = func(file, is_large_file)
        elif not vocab or not embedding:
            raise ValueError('参数"path"为空的情况下，"vocab"和"embedding"不能为空.')
        self.tokenizer = Tokenizer(vocab=vocab, special_tokens=special_tokens, lang=lang, cut_type=cut_type,
                                   keep_punctuation=keep_punctuation, cut_fn=cut_fn)
        special_tokens = self.tokenizer.special_tokens.copy()
        special_tokens.reverse()
        self.embed_dim = len(embedding[0])
        for token in special_tokens:
            embedding = [[self.tokenizer.do_encode(token)] * self.embed_dim] + embedding
        self.embedding = Embedding.from_pretrained(torch.tensor(embedding, dtype=torch.float))
        del embedding
        gc.collect()

    @property
    def vocab_size(self):
        return self.tokenizer.vocab_size

    @property
    def real_vocab(self):
        return self.tokenizer.real_vocab

    def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
               bos=False, eos=False, is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return: ([id:int], sequence_length:int) if return_sequence_length is True, else [id:int]
        """
        return self.tokenizer.encode(sentence, max_length, truncation, padding, padding_side, bos, eos,
                                     is_split_into_words, return_sequence_length)

    def batch_encode(self, sentences: Union[Collection[str], Collection[Collection]], max_length: int = None, truncation=True,
                     padding=True, padding_side='right', bos=False, eos=False,
                     is_split_into_words: bool = False, return_sequence_length=False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param is_split_into_words: 是否已经分词
        :param return_sequence_length: 是否返回序列实际长度
        :return:
            if return_sequence_length is True:
                ([[id:int]], [sequence_length:int]) if tokens is batch, else ([id:int], sequence_length:int)
            
            if return_sequence_length is False:
                [[id:int]] if tokens is batch, else [id:int]
        """
        return self.tokenizer.batch_encode(sentences, max_length, truncation, padding, padding_side, bos, eos,
                                           is_split_into_words, return_sequence_length)

    def decode(self, tokens: Collection[int], return_special_tokens=False, return_sentence=False):
        """
        :param tokens: [2, 19, 27, 3, 0, 0]
        :param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
        :param return_sentence: 返回的是一句话还是词序列
        :return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
        """
        return self.tokenizer.decode(tokens, return_special_tokens, return_sentence)

    def batch_decode(self, tokens: Collection[Collection[int]], return_special_tokens=False, return_sentence=False):
        """
        :param tokens: [[2, 19, 27, 3, 0, 0], [2, 10, 3, 0, 0, 0]]
        :param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
        :param return_sentence: 返回的是一句话还是词序列
        :return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
        """
        return self.tokenizer.batch_decode(tokens, return_special_tokens, return_sentence)

    def pad(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int, truncation=True,
            padding=True, padding_side='right', bos=False, eos=False):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :return:
        """
        return self.tokenizer.pad(tokens, max_length, truncation, padding, padding_side, bos, eos)

    def encode_plus(self, sentence: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
                    truncation=True, padding=True, padding_side='right', bos=False, eos=False, 
                    return_mask=False, is_split_into_words=False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了' 或 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param return_mask: 是否返回 mask_tokens
        :param is_split_into_words: 是否已经分词
        :return:
        """
        return self.tokenizer.encode_plus(sentence, max_length, truncation, padding, padding_side, bos, eos,
                                          return_mask, is_split_into_words)

    def batch_encode_plus(self, sentences: Union[str, Collection[str], Collection[Collection]], max_length: int = None,
                          truncation=True, padding=True, padding_side='right', bos=False, eos=False,
                          return_mask=False, is_split_into_words=False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param return_mask: 是否返回 mask_tokens
        :param is_split_into_words: 是否已经分词
        :return:
        """
        return self.tokenizer.batch_encode_plus(sentences, max_length, truncation, padding, padding_side, bos,
                                                eos, return_mask, is_split_into_words)

    def pad_plus(self, tokens: Union[map, Collection[int], Collection[Collection[int]]], max_length: int, truncation=True,
                 padding=True, padding_side='right', bos=False, eos=False, return_mask=False):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param bos:
        :param eos:
        :param return_mask:
        :return:
        """
        return self.tokenizer.pad_plus(tokens, max_length, truncation, padding, padding_side, bos, eos, return_mask)

    def __call__(self, sentence: Union[str, Collection[str], Collection[Collection]], max_length: int = None, truncation=True,
                 padding=True, padding_side='right', bos=False, eos=False, is_split_into_words: bool = False):
        input_tokens = self.batch_encode(sentence, max_length, truncation, padding, padding_side, bos, eos, is_split_into_words)
        return self.embedding(torch.tensor(input_tokens, dtype=torch.long))

    @classmethod
    def from_file(cls, file: Union[str, Path], func=load_embedding, is_large_file=False, special_tokens=[], lang='cn',
                  cut_type='word', keep_punctuation=True, cut_fn=None):
        """
        :param file: embedding文件， 如：'./sgns.weibo.word.bz2'. 注意：必须是单一文件，不能是文件夹。
        :param func: 具体读取文件的处理函数，load_embedding，可替换。注意：其函数签名为 function_name(path: str, is_large_file: bool) -> [vocab], [[embedding]]
        :param is_large_file: 是否是大文件
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        return cls(file=file, special_tokens=special_tokens, lang=lang, cut_type=cut_type, func=func,
                   is_large_file=is_large_file, keep_punctuation=keep_punctuation, cut_fn=cut_fn)

    @classmethod
    def from_vocab_embedding(cls, vocab: Collection[str], embedding: Collection[Collection[float]], large_file=False,
                             special_tokens: Collection[str] = None, lang='cn', cut_type='word', keep_punctuation=True,
                             cut_fn=None):
        """
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param embedding: [[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578]]
        :param large_file: 是否是大文件
        :param special_tokens: 保留token, 如 '<pad>', '<unk>'等
        :param lang: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param keep_punctuation: 是否保留标点符号
        :param cut_fn
        """
        return cls(vocab=vocab, embedding=embedding, large_file=large_file, special_tokens=special_tokens,
                   lang=lang, cut_type=cut_type, keep_punctuation=keep_punctuation, cut_fn=cut_fn)


# if __name__ == '__main__':
    # file_name = './article.txt'
    # text = read_file(file_name)

    # import pandas as pd
    # import matplotlib.pyplot as plt

    # DATA_PATH = r'D:\Study\kkb\代码实战课\week1\toutiao-text-classify\dataset\train.csv'
    # df = pd.read_csv(DATA_PATH)  # ['label']  ['sentence']

    # show_label_category_count(df['label'])
    # show_sentence_len_hist(df['sentence'], lang='cn', cut_type='word', bins=50, scope=(0, 30))
    # show_token_freq_plot(df['sentence'], lang='cn', cut_type='word', scope=(0, 8000))

    # tokenizer = Tokenizer(texts=df['sentence'].values, min_freq=10, lang='cn', cut_type='word')
    # tokenizer = BaseTokenizer.from_texts(df['sentence'].values)
    # tokenizer.save()

    # sent = ('上课时学生手机响个不停，老师一怒之下把手机摔了，家长拿发票让老师赔，大家怎么看待这种事？', '老师一怒之下把手机摔了')
    # tokenizer = BaseTokenizer.load()
    # d = tokenizer.encode(sent, max_length=30, keep_punctuation=True, bos=True, eos=True)
    # print(d)
    # print(tokenizer.decode(d, return_special_tokens=False, return_sentence=True))
    # tokenizer = Tokenizer.load()
    # # print(type(tokenizer))
    # # # print(len(tokenizer))
    # # # print(tokenizer.vocab[:20])
    # # # print(tokenizer.special_tokens)
    # # # print(tokenizer.encode(['the', 'my']))
    # #
    # print(tokenizer.vocab[:20])
    # print(sent)
    # e = tokenizer(sent, max_length=30, keep_punctuation=True, bos=True, eos=True)
    # print(e)
    # d = tokenizer.decode(e)
    # print(d)

    # print(tokenizer.decode(d['input_tokens'], return_special_tokens=False, return_sentence=True))
    # print(tokenizer.decode(d['input_tokens']))

    # tk = [[101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102, 7313, 4764, 3221, 4507, 102, 4764, 3221],
    #       [101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102]]
    # r = tokenizer.padding(tk, max_length=15, truncation=True, padding_side='right', bos=True, eos=True)
    # print(r)

    # PATH = r'D:\Study\kkb\代码实战课\week1\toutiao-text-classify\dataset\sgns.weibo.word.bz2'
    # tokenizer = TokenEmbedding(PATH)
    # tokenizer = TokenEmbedding.from_file(PATH, is_large_file=False)
    # d = tokenizer.encode(sent, max_length=30, keep_punctuation=True, bos=True, eos=True)
    # print(d)
    # print(tokenizer.decode(d['input_tokens'], return_special_tokens=False, return_sentence=True))
    # print(tokenizer.decode(d['input_tokens']))

    # tk = [[101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102, 7313, 4764, 3221, 4507, 102, 4764, 3221],
    #       [101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102]]
    # r = tokenizer.padding(tk, max_length=15, truncation=True, padding_side='right', bos=True, eos=True)
    # print(r)

    # print(batch_pad_mask(tk))

    # tokenizer = BaseTokenizer.from_file('./', pattern='*.txt', lang='en')
    # print(tokenizer.vocab)

    # for f in get_files(r'D:\tmp'):
    #     print(f)

    # for line in read_large_file('./article.txt'):
    #     print(line)



import os
import re
import operator
from pathlib import Path
from functools import reduce
from itertools import tee, chain
from typing import List, Union, Iterable, Collection
from concurrent.futures import as_completed, ProcessPoolExecutor

UTF8 = 'utf-8'
REMOVE_CHARS = '[·’!"\#$%&\'()＃！（）*+,-./:;<=>?\@，。：?￥★、…．＞【】［］《》？“”‘’\[\\]^_`{|}~]+'
REMOVE_CHARS_PATTERN = re.compile(REMOVE_CHARS)
EN_PATTERN = re.compile('[^A-Za-z]+')
EN_CHAR_PATTERN = re.compile(r'([\W])')
CN_CHAR_PATTERN = re.compile(r'([\u4e00-\u9fa5])')


def convert_labels(labels):
	labels = labels.astype('category').cat
	return labels.codes.to_numpy(), labels.categories.tolist()


def get_text_length(text: str, lang='cn', cut_type='word', keep_punctuation=False, cut_fn=None) -> int:
	if cut_fn is None and lang == 'cn' and cut_type == 'word':
		import jieba
		import logging
		jieba.setLogLevel(logging.INFO)
		cut_fn = jieba.lcut
	cuts = cut(text, lang, cut_type, keep_punctuation, cut_fn=cut_fn)
	return len(cuts)


def get_texts_max_length(texts: List[str], lang='cn', cut_type='word', keep_punctuation=False, cut_fn=None) -> int:
	if cut_fn is None and lang == 'cn' and cut_type == 'word':
		import jieba
		import logging
		jieba.setLogLevel(logging.INFO)
		cut_fn = jieba.lcut
	batch_cuts = batch_cut(texts, lang, cut_type, keep_punctuation, cut_fn=cut_fn)
	return max(map(lambda s: len(s), batch_cuts))


def get_df_text_labels(df, text_col: str, label_col: str, min_length: int = None, max_length: int = None):
	"""Returns (texts, labels, classes)"""
	
	if min_length or max_length:
		df['length'] = df[text_col].apply(lambda x: get_text_length(x))
	if min_length:
		df = df[df['length'] >= min_length]
	if max_length:
		df = df[df['length'] <= max_length]
	
	_labels, _classes = convert_labels(df[label_col])
	_texts = df[text_col].values
	return _texts, _labels, _classes


# -----------------------------------------------------------Read file-------------------------------------------------
def read_file(path: Union[str, Path], encoding=UTF8):
	"""
    读取文件内容
    :param path: 文件名，不能是文件夹
    :param encoding: 编码
    :return: 包含非空文本行的生成器，如 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
    """
	# if Path(path).stat().st_size <= 10000: #1048576000: # 小于等于1G
	with open(path, encoding=encoding) as f:
		lines = f.readlines()
	return [line.strip() for line in lines if line.strip()]


def read_large_file(path: Union[str, Path], encoding=UTF8):
	with open(path, encoding=encoding) as f:
		for line in f:
			line = line.strip()
			if line:
				yield line


async def async_read_file(path: Union[str, Path], encoding=UTF8):
	return read_file(path, encoding)


def read_corpus_files(path: Union[str, Path], encoding=UTF8, pattern='*', func=read_file, async_func=async_read_file):
	"""
    读取文件或文件夹下所有符合条件的文件
    :param path: 文件或文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
    :param encoding: 编码
    :param pattern: 当path是文件夹的时候，会根据此后缀过滤文件
    :param func: 具体读取文件的读函数，默认是read_file，可替换。注意：其函数签名为 function_name(path: str, encoding: str) -> corpus: Iterable[str]
    :param async_func: 传入文件夹时的读函数，用协程读取每个文件，默认是async_read_file，可替换。注意：其函数签名为 async function_name(path: str, encoding: str) -> corpus: Iterable[str]
    :return: 非空文本行，如 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
    """
	if isinstance(path, str):
		path = Path(path)
	
	if path.is_file():
		return func(path, encoding)
	
	import asyncio
	loop = asyncio.get_event_loop()
	tasks = [loop.create_task(async_func(file, encoding)) for file in path.rglob(pattern)]  # 这里不能用map，否则读不出数据
	wait_coro = asyncio.wait(tasks)
	loop.run_until_complete(wait_coro)
	all_lines = (task.result() for task in tasks)
	loop.close()
	
	return reduce(operator.iconcat, all_lines, [])


def load_embedding(path: str, is_large_file=False):
	if path.endswith('.bz2'):
		return load_embedding_from_bz2(path, is_large_file)
	return load_embedding_nomal(path, is_large_file)


def load_embedding_from_bz2(path: str, is_large_file=False):
	import bz2
	if is_large_file:
		with bz2.open(path, 'r') as f:
			tokens, vecs = _get_token_vecs(f, require_decode_token=True)
		return list(tokens), list(vecs)
	
	with bz2.open(path, 'r') as f:
		lines = f.readlines()
	return _handle_lines(lines, require_decode_token=True)


def load_embedding_nomal(path: str, is_large_file=False):
	if is_large_file:
		with open(path, 'r', encoding=UTF8) as f:
			tokens, vecs = _get_token_vecs(f, require_decode_token=False)
		return list(tokens), list(vecs)
	
	with open(path, 'r', encoding=UTF8) as f:
		lines = f.readlines()
	return _handle_lines(lines, require_decode_token=False)


def _get_token_vecs(f, require_decode_token):
	token_vec = (_handle_line(line, require_decode_token) for line in f if len(line.rstrip().split()) > 2)
	return zip(*token_vec)


def _handle_lines(lines: Iterable[str], require_decode_token: bool, batch_size=8192):
	if len(lines[0].split()) <= 2:
		lines = lines[1:]
	
	length = len(lines)
	if length <= (batch_size << 2):
		token_vecs = list(map(lambda line: _handle_line(line, require_decode_token), lines))
		tokens, vecs = zip(*token_vecs)
		return list(tokens), list(vecs)
	
	splits = [lines[i:i + batch_size] for i in range(0, length, batch_size)]
	n_jobs = min(len(splits), os.cpu_count() - 1)
	with ProcessPoolExecutor(n_jobs) as executor:
		results = [executor.submit(_order_handle_lines, i, split, require_decode_token) for i, split in enumerate(splits)]
		results = [r.result() for r in as_completed(results)]
	results.sort(key=lambda x: x[0], reverse=False)
	results = [r[1] for r in results]
	token_vecs = list(chain(*results))  # [[str,],] => [str,]
	tokens, vecs = zip(*token_vecs)
	return list(tokens), list(vecs)


def _order_handle_lines(order: int, lines: Iterable[str], require_decode_token: bool):
	return order, list(map(lambda line: _handle_line(line, require_decode_token), lines))


def _handle_line(line: str, require_decode_token: bool):
	def get_vec(elems):
		return list(map(float, elems))
	
	elems = line.rstrip().split()
	return elems[0].decode(UTF8) if require_decode_token else elems[0], get_vec(elems[1:])


# --------------------------------------------------Token cut----------------------------------------------------------
def cut_char(sentence: str):
	"""
    把句子按字分开，不破坏英文结构
    把所有的中文分开，同时希望英文和数字不能被拆分
    """
	parts = EN_CHAR_PATTERN.split(sentence)
	parts = filter(str.strip, parts)
	results = map(_cn_char_handle, parts)
	return list(chain(*results))


def batch_cut(text: Collection[str], lang='cn', cut_type='word', keep_punctuation=False, cut_fn=None):
	"""
    多句话批量分词
    :param text: 多句话，即多行
    :param lang: 哪国语言，支持cn和en
    :param cut_type: 按词还是字分，支持word和char
    :param keep_punctuation: 是否保留标点符号
    :param cut_fn
    :return: 分词后的list(2维)
    """
	if lang == 'cn':
		replace_char = ''
		if cut_type == 'word':
			if keep_punctuation:
				def fn(s):
					return cut_fn(s.strip())
			else:
				def fn(s):
					return cut_fn(re.sub(REMOVE_CHARS_PATTERN, replace_char, s.strip()))
		else:
			if cut_fn is None:
				cut_fn = cut_char
			
			if keep_punctuation:
				def fn(s):
					return cut_fn(s.strip())
			else:
				def fn(s):
					return cut_fn(re.sub(REMOVE_CHARS_PATTERN, replace_char, s.strip()))
		return map(fn, text)
		# return [cut_char(re.sub(REMOVE_CHARS_PATTERN, '', line.strip())) for line in text]
	
	if lang == 'en':
		replace_char = ' '
		if cut_type == 'word':
			def fn(s):
				return re.sub(REMOVE_CHARS_PATTERN, replace_char, s).strip().lower().split()
		else:
			if keep_punctuation:
				def fn(s):
					return list(s.strip().lower())
			else:
				def fn(s):
					return list(re.sub(EN_PATTERN, replace_char, s).strip().lower())
		return map(fn, text)
		# return [fn(line) for line in text]
	
	raise NotImplementedError(f'暂时未实现"{lang}"的分词功能')


def cut(sentence: str, lang='cn', cut_type='word', keep_punctuation=False, cut_fn=None):
	"""
    单句话分词
    :param sentence: 单句话，即一行
    :param lang: 哪国语言，支持cn和en
    :param cut_type: 按词还是字分，支持word和char
    :param keep_punctuation: 是否保留标点符号
    :param cut_fn
    :return: 分词后的list(1维)
    """
	import re
	if lang == 'cn':
		replace_char = ''
		if cut_type == 'word':
			if keep_punctuation:
				return list(cut_fn(sentence.strip()))
			else:
				return list(cut_fn(re.sub(REMOVE_CHARS_PATTERN, replace_char, sentence.strip())))
		else:
			if cut_fn is None:
				cut_fn = cut_char
			
			if keep_punctuation:
				return cut_fn(sentence.strip())
			else:
				return cut_fn(re.sub(REMOVE_CHARS_PATTERN, replace_char, sentence.strip()))
	
	if lang == 'en':
		replace_char = ' '
		if cut_type == 'word':
			return re.sub(REMOVE_CHARS_PATTERN, replace_char, sentence).strip().lower().split()
		else:
			if keep_punctuation:
				return list(sentence.strip().lower())
			else:
				return list(re.sub(EN_PATTERN, replace_char, sentence).strip().lower())
	
	raise NotImplementedError(f'暂时未实现"{lang}"的分词功能')


def filter_stop_words(cut_corpus: Iterable[List[str]], stop_words: Collection[str]):
	return map(lambda line: [word for word in line if word not in stop_words], cut_corpus)
	# return map(lambda line: list(filter(lambda word: word not in stop_words, line)), cut_corpus)


# -------------------------------------------------pad function--------------------------------------------------------
def pad(ids: List[int], max_length: int = None, truncation=True, padding=True, padding_side='right',
        bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	"""
    pad token list(1维)
    :return: padded list, valid_size size
    """
	assert padding_side in ('right', 'left'), '参数padding_side只能是"right"或"left".'
	size = len(ids)
	if bos and eos:
		size += 2
		if max_length and truncation and size > max_length:
			return [bos_id] + ids[:max_length - 2] + [eos_id], max_length
		if padding and max_length and size < max_length:
			if padding_side == 'right':
				return [bos_id] + ids + [eos_id] + [pad_id] * (max_length - size), size
			elif padding_side == 'left':
				return [pad_id] * (max_length - size) + [bos_id] + ids + [eos_id], size
			raise ValueError(f'参数"padding_side"错误: {padding_side}')
		return [bos_id] + ids + [eos_id], size
	elif bos:
		size += 1
		if max_length and truncation and size > max_length:
			return [bos_id] + ids[:max_length - 2] + [eos_id], max_length
		if padding and max_length and size < max_length:
			if padding_side == 'right':
				return [bos_id] + ids + [pad_id] * (max_length - size), size
			elif padding_side == 'left':
				return [pad_id] * (max_length - size) + [bos_id] + ids, size
			raise ValueError(f'参数"padding_side"错误: {padding_side}')
		return [bos_id] + ids, size
	elif eos:
		size += 1
		if max_length and truncation and size > max_length:
			return ids[:max_length - 2] + [eos_id], max_length
		if padding and max_length and size < max_length:
			if padding_side == 'right':
				return ids + [eos_id] + [pad_id] * (max_length - size), size
			elif padding_side == 'left':
				return [pad_id] * (max_length - size) + ids + [eos_id], size
			raise ValueError(f'参数"padding_side"错误: {padding_side}')
		return ids + [eos_id], size
	else:
		if max_length and truncation and size > max_length:
			return ids[:max_length], max_length
		if padding and max_length and size < max_length:
			if padding_side == 'right':
				return ids + [pad_id] * (max_length - size), size
			elif padding_side == 'left':
				return [pad_id] * (max_length - size) + ids, size
			raise ValueError(f'参数"padding_side"错误: {padding_side}')
		return ids, size


def batch_pad(batch: Union[map, Iterable[Iterable[int]]], max_length: int = None, truncation=True,
              padding=True, padding_side='right', bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	"""
    :param batch: (2维)
    :param max_length:
    :param truncation:
    :param padding_side:
    :param bos:
    :param eos:
    :param pad_id:
    :param bos_id:
    :param eos_id:
    :return:
    """
	if not max_length and (truncation or padding):
		batch, batch_copy = tee(batch, 2)
		max_length = max(map(len, batch_copy))
		if bos:
			max_length += 1
		if eos:
			max_length += 1
	ids, sizes = zip(*map(lambda ids: pad(ids, max_length, truncation, padding, padding_side, bos, eos, pad_id, bos_id, eos_id),batch))
	return list(ids), list(sizes)


def _pad_mask(ids: Iterable[int], max_length: int = None, truncation=True, padding=True, padding_side='right',
              bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	input_ids, real_size = pad(ids, max_length, truncation, padding, padding_side, bos, eos, pad_id, bos_id, eos_id)
	size = len(input_ids)
	if size > real_size:
		if padding_side == 'right':
			return input_ids, [1] * real_size + [0] * (size - real_size)
		return input_ids, [0] * (size - real_size) + [1] * real_size
	else:
		return input_ids, [1] * size


def pad_mask(ids: Iterable[int], max_length: int = None, truncation=True, padding=True, padding_side='right',
             bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	input_ids, mask_ids = _pad_mask(ids, max_length, truncation, padding, padding_side, bos, eos, pad_id, bos_id,
	                                eos_id)
	return {'input_ids': input_ids, 'mask_ids': mask_ids}


def batch_pad_mask(batch: Union[map, Iterable[Iterable[int]]], max_length: int = None, truncation=True,
                   padding=True, padding_side='right', bos=False, eos=False, pad_id=0, bos_id=2, eos_id=3):
	if not max_length and (truncation or padding):
		batch, batch_copy = tee(batch, 2)
		max_length = max(map(len, batch_copy))
		if bos:
			max_length += 1
		if eos:
			max_length += 1
	ids = list(
		map(lambda ids: _pad_mask(ids, max_length, truncation, padding, padding_side, bos, eos, pad_id, bos_id, eos_id),
		    batch))
	input_ids, mask_ids = zip(*ids)
	return {'input_ids': list(input_ids), 'mask_ids': list(mask_ids)}


def _cn_char_handle(part):
	chars = CN_CHAR_PATTERN.split(part)
	return filter(str.strip, chars)



from tqdm import tqdm
from copy import deepcopy
import numpy as np
from typing import List, Tuple, Union
import torch
from torch import optim
from torch.utils.data import DataLoader, TensorDataset, Dataset
from torch.optim.lr_scheduler import LRScheduler, CosineAnnealingLR
from sklearn.model_selection import train_test_split
from nlpx import log_utils


def cal_count(y):
	shape = y.shape
	return shape[0] if len(shape) == 1 else shape[0] * shape[1]


def cal_correct(logits: torch.Tensor, y: torch.Tensor):
	if len(logits.size()) > 1:
		return (logits.argmax(-1) == y).sum()
	else:
		return (logits > 0.5).sum()


def evaluate(model, val_loader, device):
	total, steps = 0, 0
	total_loss, correct = [torch.Tensor([0.0]) for _ in range(2)]
	model.eval()
	with torch.no_grad():
		if hasattr(model, 'fit'):
			for batch in val_loader:
				batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
				y = batch[-1]
				logits, loss = model.fit(*batch)
				total_loss += loss
				total += cal_count(y)
				correct += cal_correct(logits, y)
				steps += 1
		else:
			for batch in val_loader:
				batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
				y = batch[-1]
				logits, loss = model(*batch)
				total_loss += loss
				total += cal_count(y)
				correct += cal_correct(logits, y)
				steps += 1
	return (total_loss.item() / steps), (correct.item() / total)


def do_train(model, batch, optimizer, device):
	batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
	logits, loss = model(*batch)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss


def do_fit(model, batch, optimizer, device):
	batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
	logits, loss = model.fit(*batch)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss


def do_train_scheduler(model, batch, optimizer, device, scheduler: LRScheduler):
	batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
	logits, loss = model(*batch)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss


def do_fit_scheduler(model, batch, optimizer, device, scheduler: LRScheduler):
	batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
	logits, loss = model.fit(*batch)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss


def do_train_acc(model, batch, optimizer, device):
	batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
	y = batch[-1]
	logits, loss = model(*batch)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss, cal_count(y), cal_correct(logits, y)


def do_fit_acc(model, batch, optimizer, device):
	batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
	y = batch[-1]
	logits, loss = model.fit(*batch)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	return loss, cal_count(y), cal_correct(logits, y)


def do_train_scheduler_acc(model, batch, optimizer, device, scheduler: LRScheduler):
	batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
	y = batch[-1]
	logits, loss = model(*batch)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss, cal_count(y), cal_correct(logits, y)


def do_fit_scheduler_acc(model, batch, optimizer, device, scheduler: LRScheduler):
	batch = [x.to(device) if torch.is_tensor(x) else x for x in batch]
	y = batch[-1]
	logits, loss = model.fit(*batch)
	optimizer.zero_grad()
	loss.backward()
	optimizer.step()
	scheduler.step()
	return loss, cal_count(y), cal_correct(logits, y)


def train_epoch_base(model, train_loader, optimizer, device):
	steps = 0
	total_loss = torch.Tensor([0.0])
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss = do_fit(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
	else:
		for batch in train_loader:
			loss = do_train(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
	
	return total_loss.item() / steps


def train_epoch_progress(model, train_loader, optimizer, device, epoch, max_iter):
	steps = 0
	total_loss = torch.Tensor([0.0])
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{max_iter}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss = do_fit(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
			loop.set_postfix(loss=f"{total_loss.item() / steps:.4f}", lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss = do_train(model, batch, optimizer, device)
			total_loss += loss
			steps += 1
			loop.set_postfix(loss=f"{total_loss.item() / steps:.4f}", lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	return total_loss.item() / steps


def train_epoch_scheduler(model, train_loader, optimizer, device, scheduler: LRScheduler):
	steps = 0
	total_loss = torch.Tensor([0.0])
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss = do_fit_scheduler(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
	else:
		for batch in train_loader:
			loss = do_train_scheduler(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
	return total_loss.item() / steps


def train_epoch_scheduler_progress(model, train_loader, optimizer, device, scheduler, epoch, max_iter):
	steps = 0
	total_loss = torch.Tensor([0.0])
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{max_iter}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss = do_fit_scheduler(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
			loop.set_postfix(loss=f"{total_loss.item() / steps:.4f}", lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss = do_train_scheduler(model, batch, optimizer, device, scheduler)
			total_loss += loss
			steps += 1
			loop.set_postfix(loss=f"{total_loss.item() / steps:.4f}", lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	return total_loss.item() / steps


def train_epoch(model, train_loader, optimizer, device, scheduler, epoch, max_iter, show_progress):
	if show_progress:
		if scheduler is None:
			return train_epoch_progress(model, train_loader, optimizer, device, epoch, max_iter)
		return train_epoch_scheduler_progress(model, train_loader, optimizer, device, scheduler, epoch, max_iter)
	else:
		if scheduler is None:
			return train_epoch_base(model, train_loader, optimizer, device)
		return train_epoch_scheduler(model, train_loader, optimizer, device, scheduler)


def train_epoch_base_acc(model, train_loader, optimizer, device):
	total, steps = 0, 0
	total_loss, total_correct = [torch.Tensor([0.0]) for _ in range(2)]
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss, count, correct = do_fit_acc(model, batch, optimizer, device)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
	else:
		for batch in train_loader:
			loss, count, correct = do_train_acc(model, batch, optimizer, device)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
	return total_correct.item() / total, total_loss.item() / steps


def train_epoch_progress_acc(model, train_loader, optimizer, device, epoch, max_iter):
	total, steps = 0, 0
	total_loss, total_correct = [torch.Tensor([0.0]) for _ in range(2)]
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{max_iter}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss, count, correct = do_fit_acc(model, batch, optimizer, device)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
			loop.set_postfix(acc=f"{total_correct.item() / total:.4f}", loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss, count, correct = do_train_acc(model, batch, optimizer, device)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
			loop.set_postfix(acc=f"{total_correct.item() / total:.4f}", loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	
	return total_correct.item() / total, total_loss.item() / steps


def train_epoch_scheduler_acc(model, train_loader, optimizer, device, scheduler: LRScheduler):
	total, steps = 0, 0
	total_loss, total_correct = [torch.Tensor([0.0]) for _ in range(2)]
	if hasattr(model, 'fit'):
		for batch in train_loader:
			loss, count, correct = do_fit_scheduler_acc(model, batch, optimizer, device, scheduler)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
	else:
		for batch in train_loader:
			loss, count, correct = do_train_scheduler_acc(model, batch, optimizer, device, scheduler)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
	return total_correct.item() / total, total_loss.item() / steps


def train_epoch_scheduler_progress_acc(model, train_loader, optimizer, device, scheduler, epoch, max_iter):
	total, steps = 0, 0
	total_loss, total_correct = [torch.Tensor([0.0]) for _ in range(2)]
	loop = tqdm(train_loader, desc=f"[Epoch-{epoch}/{max_iter}]", total=len(train_loader), colour="green")
	if hasattr(model, 'fit'):
		for batch in loop:
			loss, count, correct = do_fit_scheduler_acc(model, batch, optimizer, device, scheduler)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
			loop.set_postfix(acc=f"{total_correct.item() / total:.4f}", loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	else:
		for batch in loop:
			loss, count, correct = do_train_scheduler_acc(model, batch, optimizer, device, scheduler)
			total_loss += loss
			total += count
			total_correct += correct
			steps += 1
			loop.set_postfix(acc=f"{total_correct.item() / total:.4f}", loss=f"{total_loss.item() / steps:.4f}",
			                 lr=f'{optimizer.param_groups[0]["lr"]:.6f}')
	loop.close()
	
	return total_correct.item() / total, total_loss.item() / steps


def train_epoch_acc(model, train_loader, optimizer, device, scheduler, epoch, max_iter, show_progress):
	if show_progress:
		if scheduler is None:
			return train_epoch_progress_acc(model, train_loader, optimizer, device, epoch, max_iter)
		return train_epoch_scheduler_progress_acc(model, train_loader, optimizer, device, scheduler, epoch, max_iter)
	else:
		if scheduler is None:
			return train_epoch_base_acc(model, train_loader, optimizer, device)
		return train_epoch_scheduler_acc(model, train_loader, optimizer, device, scheduler)


class Trainer:
	
	def __init__(self, max_iter=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             learning_rate=0.001, T_max: int = 0, batch_size=32, num_workers=0,
	             pin_memory: bool = False, pin_memory_device: str = "", persistent_workers: bool = False,
	             early_stopping_rounds: int = 2,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt', device=torch.device
	             ):
		self.max_iter = max_iter
		self.optimizer = optimizer
		self.scheduler = scheduler
		self.learning_rate = learning_rate
		self.T_max = T_max
		self.batch_size = batch_size
		self.num_workers = num_workers
		self.pin_memory = pin_memory
		self.pin_memory_device = pin_memory_device
		self.persistent_workers = persistent_workers
		self.early_stopping_rounds = early_stopping_rounds
		self.print_per_rounds = print_per_rounds
		self.checkpoint_per_rounds = checkpoint_per_rounds
		self.checkpoint_name = checkpoint_name
		self.device = device or torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
	
	def train(self, model, train_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		cnt = 0
		min_loss = float('inf')
		best_model = None
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		model.train()
		for epoch in range(1, self.max_iter + 1):
			avg_loss = train_epoch(model, train_loader, optimizer, self.device, scheduler, epoch, self.max_iter,
			                       show_progress)
			
			self.try_print(self.print, epoch, optimizer.param_groups[0]["lr"], avg_loss)
			self.try_checkpoint(model, epoch)
			
			if min_loss - avg_loss > eps:
				cnt = 0
				best_model = deepcopy(model)
				min_loss = avg_loss
				continue
			
			# x次epoch的val_acc不提升或x次epoch的val_acc不变化
			if epoch >= min(3, self.early_stopping_rounds) and cnt >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.max_iter}")
				break
			
			cnt += 1
		
		return best_model
	
	def get_optimizer_scheduler(self, model):
		scheduler = None
		if self.scheduler is not None:
			scheduler = self.scheduler
			optimizer = scheduler.optimizer
		elif self.optimizer is None:
			optimizer = optim.AdamW(model.parameters(), lr=self.learning_rate)
		elif isinstance(self.optimizer, type):
			optimizer = self.optimizer(model.parameters(), lr=self.learning_rate)
		else:
			optimizer = self.optimizer
		
		if scheduler is None and self.T_max and self.T_max > 0:
			scheduler = CosineAnnealingLR(optimizer, T_max=self.T_max)
		
		return optimizer, scheduler
	
	def try_checkpoint(self, model, epoch):
		if self.checkpoint_per_rounds <= 0:
			return
		
		if self.checkpoint_per_rounds == 1 or epoch % self.checkpoint_per_rounds == 0:
			torch.save(model, self.checkpoint_name)
			
	def try_print(self, do_print, epoch, lr, loss, **kwargs):
		if self.print_per_rounds == 1 or epoch % self.print_per_rounds == 0:
			do_print(epoch, lr, loss, **kwargs)
			
	def print(self, epoch, lr, loss):
		log_utils.info(f'Epoch-{epoch}/{self.max_iter}  lr: {lr:.6f}, loss: {loss:.4f}')


class SimpleTrainer(Trainer):
	
	def __init__(self, max_iter=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             learning_rate=0.001, T_max: int = 0, batch_size=32, num_workers=1,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds=10,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt', device=torch.device):
		super().__init__(max_iter, optimizer, scheduler, learning_rate, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds, 
		                 checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, collate_fn=None, show_progress=False, eps=1e-5):
		if isinstance(X, (List, np.ndarray)):
			X = torch.tensor(X, dtype=torch.float)
		if isinstance(y, (List, np.ndarray)):
			y = torch.tensor(y, dtype=torch.long)
		
		return super().train(model, TensorDataset(X, y), collate_fn, show_progress, eps=eps)


class ClassTrainer(Trainer):
	
	def __init__(self, max_iter=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             learning_rate=0.001, T_max: int = 0, batch_size=32, num_workers=1,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds=10,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt', device=torch.device):
		super().__init__(max_iter, optimizer, scheduler, learning_rate, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds, 
		                 checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, train_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		best_model = None
		min_loss = float('inf')
		cnt, cnt2, best_acc, last_acc = 0, 0, 0.0, 0.0
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		model.train()
		for epoch in range(1, self.max_iter + 1):
			acc, avg_loss = train_epoch_acc(model, train_loader, optimizer, self.device, scheduler, epoch,
			                                self.max_iter, show_progress)
			
			self.try_print(self.print, epoch, optimizer.param_groups[0]["lr"], avg_loss, acc=acc)
			self.try_checkpoint(model, epoch)
			
			if acc > best_acc or (acc == best_acc and min_loss - avg_loss > eps):
				cnt, cnt2 = 0, 0
				best_acc, best_model = acc, deepcopy(model)
				last_acc, min_loss = acc, min(min_loss, avg_loss)
				continue
			
			# x次epoch的val_acc不提升或x次epoch的val_acc不变化
			if epoch >= min(5, self.early_stopping_rounds) and max(cnt, cnt2) >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.max_iter}")
				break
			
			if abs(last_acc - acc) < eps:  # val_acc不在变化
				cnt2 += 1
			else:
				cnt2 = 0
			
			cnt += 1
			last_acc = acc
			best_acc = max(best_acc, acc)
			min_loss = min(min_loss, avg_loss)
		
		return best_model
	
	def print(self, epoch, lr, loss, acc):
		log_utils.info(f'Epoch-{epoch}/{self.max_iter}  lr: {lr:.6f}, train_loss: {loss:.4f}, train_acc: {acc:.4f}')


class SimpleClassTrainer(ClassTrainer):
	
	def __init__(self, max_iter=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             learning_rate=0.001, T_max: int = 0, batch_size=32, num_workers=1,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds=10,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt', device=torch.device):
		super().__init__(max_iter, optimizer, scheduler, learning_rate, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, collate_fn=None, show_progress=False, eps=1e-5):
		if isinstance(X, (List, np.ndarray)):
			X = torch.tensor(X, dtype=torch.float)
		if isinstance(y, (List, np.ndarray)):
			y = torch.tensor(y, dtype=torch.long)
		
		return super().train(model, TensorDataset(X, y), collate_fn, show_progress, eps=eps)


class EvalClassTrainer(Trainer):
	
	def __init__(self, max_iter=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             learning_rate=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0, pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False, early_stopping_rounds=10,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt', device=torch.device):
		super().__init__(max_iter, optimizer, scheduler, learning_rate, T_max, batch_size, num_workers, pin_memory,
		                 pin_memory_device, persistent_workers, early_stopping_rounds, print_per_rounds,
		                 checkpoint_per_rounds, checkpoint_name, device)
		self.eval_batch_size = eval_batch_size
		self.num_eval_workers = num_eval_workers
	
	def train(self, model, train_set: Dataset, val_set: Dataset, collate_fn=None, show_progress=False, eps=1e-5):
		train_loader = DataLoader(dataset=train_set, batch_size=self.batch_size, pin_memory=self.pin_memory,
		                          pin_memory_device=self.pin_memory_device, persistent_workers=self.persistent_workers,
		                          num_workers=self.num_workers, shuffle=True, collate_fn=collate_fn)
		val_loader = DataLoader(dataset=val_set, batch_size=self.eval_batch_size,
		                        num_workers=self.num_eval_workers, collate_fn=collate_fn)
		best_model = None
		min_loss = float('inf')
		cnt, cnt2, best_acc, last_acc = 0, 0, 0.0, 0.0
		optimizer, scheduler = self.get_optimizer_scheduler(model)
		
		for epoch in range(1, self.max_iter + 1):
			model.train()
			acc, avg_loss = train_epoch_acc(model, train_loader, optimizer, self.device, scheduler, epoch,
			                                self.max_iter, show_progress)
			
			val_loss, val_acc = evaluate(model, val_loader, self.device)
			
			self.try_print(self.print, epoch, optimizer.param_groups[0]["lr"], avg_loss,
			           acc=acc, val_loss=val_loss, val_acc=val_acc)
			self.try_checkpoint(model, epoch)
			
			if val_acc > best_acc or (val_acc == best_acc and min_loss - val_loss > eps):
				cnt, cnt2 = 0, 0
				best_acc, best_model = val_acc, deepcopy(model)
				last_acc, min_loss = val_acc, min(min_loss, val_loss)
				continue
			
			# x次epoch的val_acc不提升或x次epoch的val_acc不变化
			if epoch >= min(5, self.early_stopping_rounds) and max(cnt, cnt2) >= self.early_stopping_rounds:
				log_utils.info(f"Early stopping at Epoch-{epoch}/{self.max_iter}")
				break
			
			if abs(last_acc - val_acc) < eps:  # val_acc不在变化
				cnt2 += 1
			else:
				cnt2 = 0
			
			cnt += 1
			last_acc = val_acc
			best_acc = max(best_acc, val_acc)
			min_loss = min(min_loss, val_loss)
		
		return best_model
	
	def print(self, epoch, lr, loss, acc, val_loss, val_acc):
		log_utils.info(
			f'Epoch-{epoch}/{self.max_iter}  lr: {lr:.6f}, train_loss: {loss:.4f}, train_acc: {acc:.4f}, '
			f'val_loss: {val_loss:.4f}, val_acc: {val_acc:.4f}')


class EvalSimpleClassTrainer(EvalClassTrainer):
	
	def __init__(self, max_iter=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             learning_rate=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds=10,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt', device=torch.device):

		super().__init__(max_iter, optimizer, scheduler, learning_rate, T_max, batch_size, eval_batch_size, num_workers,
		                 num_eval_workers, pin_memory, pin_memory_device, persistent_workers, early_stopping_rounds,
		                 print_per_rounds, checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, val_data: Tuple, collate_fn=None, show_progress=False, eps=1e-5):
		if isinstance(X, (List, np.ndarray)):
			X = torch.tensor(X, dtype=torch.float)
		if isinstance(y, (List, np.ndarray)):
			y = torch.tensor(y, dtype=torch.long)
		X_val, y_val = val_data[0], val_data[1]
		if isinstance(X_val, (List, np.ndarray)):
			X_val = torch.tensor(X_val, dtype=torch.float)
		if isinstance(y_val, (List, np.ndarray)):
			y_val = torch.tensor(y_val, dtype=torch.long)
		
		return super().train(model, TensorDataset(X, y), TensorDataset(X_val, y_val), collate_fn, show_progress, eps)


class SplitClassTrainer(EvalSimpleClassTrainer):
	
	def __init__(self, max_iter=100, optimizer: Union[type, optim.Optimizer] = None, scheduler: LRScheduler = None,
	             learning_rate=0.001, T_max: int = 0, batch_size=32, eval_batch_size=64,
	             num_workers=0, num_eval_workers=0,
	             pin_memory: bool = False, pin_memory_device: str = "",
	             persistent_workers: bool = False,
	             early_stopping_rounds=10,  # 早停，等10轮决策，评价指标不在变化，停止
	             print_per_rounds: int = 1, checkpoint_per_rounds: int = 0, checkpoint_name='model.pt', device=torch.device):
		super().__init__(max_iter, optimizer, scheduler, learning_rate, T_max, batch_size, eval_batch_size, num_workers,
		                 num_eval_workers, pin_memory, pin_memory_device, persistent_workers, early_stopping_rounds,
		                 print_per_rounds, checkpoint_per_rounds, checkpoint_name, device)
	
	def train(self, model, X, y, val_size=0.2, random_state=None, collate_fn=None, show_progress=False, eps=1e-5):
		X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=val_size, random_state=random_state)
		return super().train(model, X_train, y_train, (X_test, y_test), collate_fn, show_progress, eps)
